"""Handler for postgresql:// urls."""
from typing import Tuple

from tentaclio.clients import postgres_client
from tentaclio.protocols import ReaderClosable, WriterClosable
from tentaclio.streams import csv_db_stream
from tentaclio.urls import URL


__all__ = ["PostgresURLHandler"]


def _split_url(url: URL) -> Tuple[URL, str]:
    parts = url.path.split("::")
    if len(parts) != 2 or parts[1] == "":
        raise ValueError("No table provided for dumping the csv data")
    path = parts[0]
    table = parts[1]
    conn_url = URL.from_components(
        scheme=url.scheme,
        username=url.username,
        password=url.password,
        hostname=url.hostname,
        port=url.port,
        path=path,
        query=url.query,
    )
    return conn_url, table


class PostgresURLHandler:
    """Handler for opening writers and readers ."""

    def open_reader_for(self, url: URL, mode: str, extras: dict) -> ReaderClosable:
        """Not implemented."""
        raise NotImplementedError("Reading csv from postgres not implemented")

    def open_writer_for(self, url: URL, mode: str, extras: dict) -> WriterClosable:
        """Open writer to dump the csv into a table in database.

        This stream writer only works for csv data at the moment. The url must contain
        a table as the last part of the path separated by a doble colon i.e.
        `postgresql://user:pass@host/database::table`. The target table needs the columns
        with the same names as the field names.
        """
        conn_url, table = _split_url(url)
        client = postgres_client.PostgresClient(conn_url, **extras)
        return csv_db_stream.DatabaseCsvWriter(client, table)



"""Stream handler registry to open readers and writers to urls."""
import abc
import logging
from typing import Callable, ClassVar, Iterable

from typing_extensions import Protocol

from tentaclio import clients
from tentaclio.registry import URLHandlerRegistry


logger = logging.getLogger(__name__)

__all__ = ["DB_REGISTRY", "Db"]


class Db(Protocol):
    """Interface for query-based connections."""

    # The context manager methods are included as we can't
    # inherit from typing_extensions.ContextManager and Protocol
    # at the same time 🤷
    def __enter__(self) -> "Db":
        """Enter the the context manager."""
        ...

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the the context manager."""
        ...

    # Query methods:

    @abc.abstractmethod
    def execute(self, sql_query: str, **params) -> None:
        """Execute a query against the underlying client."""
        ...

    @abc.abstractmethod
    def query(self, sql_query: str, **params) -> Iterable:
        """Perform the query and return an iterable of the results."""
        ...


DbFactory = Callable[..., Db]


class DbRegistry(URLHandlerRegistry[DbFactory]):
    """Registry for databases."""

    def get_handler(self, scheme: str) -> DbFactory:
        """Get the handler for the give scheme. Raise an URLError if no handler is registred."""
        if scheme not in self.registry:
            logger.info("trying to return the default SQLA client")
            return clients.SQLAlchemyClient

        return self.registry[scheme]


class _DbRegistryHolder:
    """Module level singleton."""

    instance: ClassVar[DbRegistry] = DbRegistry()


DB_REGISTRY = _DbRegistryHolder().instance



"""AWS Athena query client.

Overrides the `get_df` convenience methods for loading a DataFrame using PandasCursor,
which is more performant than using sql alchemy functions.
"""

import pandas as pd
from pyathena.pandas_cursor import PandasCursor

from . import decorators, sqla_client


__all__ = ["AthenaClient"]


class AthenaClient(sqla_client.SQLAlchemyClient):
    """Postgres client, backed by a SQLAlchemy connection."""

    allowed_schemes = ["awsathena+rest"]
    connect_args_default = dict(cursor_class=PandasCursor)
    # Athena-specific fast query result retrieval:

    @decorators.check_conn
    def get_df(self, sql_query: str, params: dict = None, **kwargs) -> pd.DataFrame:
        """Run a raw SQL query and return a data frame."""
        raw_conn = self._get_raw_conn()
        raw_cursor = raw_conn.cursor(PandasCursor)
        return raw_cursor.execute(sql_query, parameters=params, **kwargs).as_pandas()



"""GS Stream client."""
from typing import Optional, Union, Tuple, cast

from google.cloud import storage

from tentaclio import urls, protocols

from . import base_client, decorators, exceptions


__all__ = ["GSClient"]


class GSClient(base_client.BaseClient["GSClient"]):
    """GS client.

    Ref: https://googleapis.dev/python/storage/latest/index.html
    """

    allowed_schemes = ["gs", "gcs"]

    conn: storage.Client
    bucket: Optional[str]
    key_name: Optional[str]

    def __init__(
        self, url: Union[urls.URL, str]
    ) -> None:
        """Create a new GS client.

        The client is created based on a URL of the form
        gs://bucket/key.

        Authentication and client configuration is recommended with environment variables.
        Quoted from docs:
            "Setting the GOOGLE_APPLICATION_CREDENTIALS and GOOGLE_CLOUD_PROJECT environment
            variables will override the automatically configured credentials."

        See documentation for more information.
        https://googleapis.dev/python/google-api-core/latest/auth.html
        """
        super().__init__(url)

        self.bucket = self.url.hostname or None
        self.key_name = self.url.path[1:] if self.url.path != "" else ""

        if self.bucket == "gs":
            self.bucket = None

    def _connect(self) -> storage.Client:
        return storage.Client()

    def close(self) -> None:
        """Close the connection.

        This is fake. GS doesn't allow to close gs connections.
        """
        # gs doesn't have close method
        if self.closed:
            raise ValueError("Trying to close a closed client")
        self.closed = True

    # Stream methods:

    @decorators.check_conn
    def get(
        self,
        writer: protocols.ByteWriter,
        bucket_name: Optional[str] = None,
        key_name: Optional[str] = None
    ) -> None:
        """Download the contents from gs and write them in the provided writer.

        Arguments:
            :bucket_name: If not provided in the url at construction time.
            :key_name: If not provided in the url at construction time.

        Raises:
            GSError: If a bucket or a key is not found in the given URL or args then
            Google Cloud Exceptions: if the client raises them.

        """
        gs_bucket, gs_key = self._fetch_bucket_and_key(bucket_name, key_name)

        self._get(writer, gs_bucket, gs_key)

    @decorators.check_conn
    def put(
        self,
        reader: protocols.ByteReader,
        bucket_name: Optional[str] = None,
        key_name: Optional[str] = None
    ) -> None:
        """Up the contents to gs.

        Arguments:
            :bucket_name: If not provided in the url at construction time.
            :key_name: If not provided in the url at construction time.

        Raises:
            GSError: If a bucket or a key is not found in the given URL or args then
            Google Cloud Exceptions: if the client raises them.

        """
        gs_bucket, gs_key = self._fetch_bucket_and_key(bucket_name, key_name)

        self._put(reader, gs_bucket, gs_key)

    # Non-streaming public methods
    @decorators.check_conn
    def remove(self, bucket_name: Optional[str] = None, key_name: Optional[str] = None) -> None:
        """Remove the object from gs bucket.

        Arguments:
            :bucket_name: If not provided in the url at construction time.
            :key_name: If not provided in the url at construction time.

        Raises:
            GSError: If a bucket or a key is not found in the given URL or args then
            Google Cloud Exceptions: if the client raises them.

        """
        gs_bucket, gs_key_name = self._fetch_bucket_and_key(bucket_name, key_name)
        self._remove(gs_bucket, gs_key_name)

    # Helpers:

    def _fetch_bucket_and_key(
        self, bucket: Optional[str] = None, key: Optional[str] = None
    ) -> Tuple[str, str]:
        bucket_name = bucket or self.bucket
        if bucket_name is None:
            raise exceptions.GSError("Missing remote bucket")

        key_name = key or self.key_name
        if key_name == "":
            raise exceptions.GSError("Missing remote key")

        return bucket_name, cast(str, key_name)

    def _get_blob(self, bucket_name: str, key_name: str) -> storage.Bucket:
        """Get the bucket client."""
        bucket = self.conn.bucket(bucket_name)
        return bucket.blob(key_name)

    def _get(
        self, writer: protocols.ByteWriter, bucket_name: str, key_name: str
    ) -> None:
        """Download file on the client."""
        blob = self._get_blob(bucket_name, key_name)
        blob.download_to_file(writer)

    def _put(
        self, reader: protocols.ByteReader, bucket_name: str, key_name: str
    ) -> None:
        """Upload on the client."""
        blob = self._get_blob(bucket_name, key_name)
        blob.upload_from_file(reader)

    def _remove(
        self, bucket_name: str, key_name: str
    ) -> None:
        """Delete on the client."""
        blob = self._get_blob(bucket_name, key_name)
        blob.delete()



"""S3 Stream client."""
from typing import Generator, Iterable, Optional, Tuple, Union, cast

import boto3
from botocore import client as boto_client

from tentaclio import fs, protocols, urls

from . import base_client, decorators, exceptions


__all__ = ["S3Client"]


class S3Client(base_client.BaseClient["S3Client"]):
    """S3 client, backed by boto3.

    Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html
    """

    allowed_schemes = ["s3"]

    conn: boto_client.BaseClient
    aws_profile: Optional[str]
    conn_encrypt: bool
    aws_access_key_id: Optional[str]
    aws_secret_access_key: Optional[str]
    bucket: Optional[str]
    key_name: Optional[str]

    def __init__(
        self, url: Union[urls.URL, str], aws_profile: str = None, conn_encrypt: bool = False
    ) -> None:
        """Create a new S3 client.

        The client is created based based on a url of the form
        s3://access_key_id:secret_access_key@bucket/key.
        If the access key and the secret are not provided, the authentication
        will be delegated to boto.
        """
        self.aws_profile = aws_profile
        self.conn_encrypt = conn_encrypt
        super().__init__(url)

        self.aws_access_key_id = self.url.username or None
        self.aws_secret_access_key = self.url.password or None
        self.bucket = self.url.hostname or None
        self.key_name = self.url.path[1:] if self.url.path != "" else ""

        if self.bucket == "s3":
            self.bucket = None

    def _connect(self) -> boto_client.BaseClient:
        # Revert to Boto default credentials

        # Exception: 's3' hostname not a valid bucket

        session = boto3.session.Session(
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            profile_name=self.aws_profile,
        )
        return session.client("s3")

    def close(self) -> None:
        """Close the connection.

        This is fake. Boto doesn't allow to close s3 connections.
        """
        # s3 doesn't have close method
        if self.closed:
            raise ValueError("Trying to close a closed client")
        self.closed = True

    # Stream methods:

    @decorators.check_conn
    def get(
        self, writer: protocols.ByteWriter, bucket_name: str = None, key_name: str = None
    ) -> None:
        """Download the contents from s3 and write them in the provided writer.

        Arguments:
            :bucket_name: If not provided in the url at construction time.
            :key_name: If not provided in the url at construction time.
        """
        s3_bucket, s3_key = self._fetch_bucket_and_key(bucket_name, key_name)

        if not self._isfile(s3_bucket, s3_key):
            raise exceptions.S3Error("Unable to fetch the remote file")

        self.conn.download_fileobj(s3_bucket, s3_key, writer)

    @decorators.check_conn
    def put(
        self, reader: protocols.ByteReader, bucket_name: str = None, key_name: str = None
    ) -> None:
        """Up the contents of the reader to s3.

        Arguments:
            :bucket_name: If not provided in the url at construction time.
            :key_name: If not provided in the url at construction time.
        """
        s3_bucket, s3_key = self._fetch_bucket_and_key(bucket_name, key_name)

        extra_args = {}
        if self.conn_encrypt:
            extra_args["ServerSideEncryption"] = "AES256"

        self.conn.upload_fileobj(reader, s3_bucket, s3_key, ExtraArgs=extra_args)

    # Helpers:

    def _fetch_bucket_and_key(self, bucket: Optional[str], key: Optional[str]) -> Tuple[str, str]:
        bucket_name = bucket or self.bucket
        if bucket_name is None:
            raise exceptions.S3Error("Missing remote bucket")

        key_name = key or self.key_name
        if key_name == "":
            raise exceptions.S3Error("Missing remote key")

        return bucket_name, cast(str, key_name)

    def _isfile(self, bucket: str, key: str) -> bool:
        """Check if a key exists in a bucket."""
        try:
            self.conn.head_object(Bucket=bucket, Key=key)
            return True
        except Exception:
            return False

    # scandir related methods

    def scandir(self, **kwargs) -> Iterable[fs.DirEntry]:
        """Scan the connection url to create dir entries."""
        if self._is_root():
            return self._build_bucket_entries()
        return _KeyLister(self)

    def _build_bucket_entries(self) -> Iterable[fs.DirEntry]:
        return (
            fs.DirEntry(url=urls.URL("s3://" + name), is_dir=True, is_file=False)
            for name in self._get_buckets()
        )

    def _get_buckets(self):
        resp = self.conn.list_buckets()
        buckets = []
        for metadata in resp["Buckets"]:
            buckets.append(metadata["Name"])
        return buckets

    def _is_root(self):
        return not self.bucket

    # Copy methods

    def copy(self, source: urls.URL, dest: urls.URL):
        """Copy source into dest directly in S3."""
        source_client = S3Client(source)
        dest_client = S3Client(dest)
        with source_client:
            source_client.conn.copy(
                CopySource={"Bucket": source_client.bucket, "Key": source_client.key_name},
                Bucket=dest_client.bucket,
                Key=dest_client.key_name,
            )

    @decorators.check_conn
    def remove(self):
        """Remove the key from the aws bucket."""
        self.conn.delete_object(Bucket=self.bucket, Key=self.key_name)


class _KeyLister(Iterable[fs.DirEntry]):
    """List S3 keys.

    Heavily inspired by
    https://github.com/dask/s3fs/blob/05fc3d9c64a5f9ff701f9d9541f707aff5750349/s3fs/core.py#L357
    """

    def __init__(self, client: S3Client):
        self.paginator = client.conn.get_paginator("list_objects_v2")
        self.bucket = client.bucket
        self.key = client.key_name

    def _get_delimeted_key(self):
        if self.key and self.key[-1] != "/":
            return self.key + "/"
        return self.key

    def __iter__(self) -> Generator[fs.DirEntry, None, None]:
        for page in self.paginator.paginate(
            Bucket=self.bucket, Prefix=self._get_delimeted_key(), Delimiter="/"
        ):
            yield from self._iter_page(page)

    def _iter_page(self, page) -> Generator[fs.DirEntry, None, None]:
        dirs = [entry["Prefix"] for entry in page.get("CommonPrefixes", [])]
        files = [entry["Key"] for entry in page.get("Contents", [])]
        for dir_ in dirs:
            yield fs.build_folder_entry(_build_url(str(self.bucket), dir_))

        for file_ in files:
            yield fs.build_file_entry(_build_url(str(self.bucket), file_))


def _build_url(bucket: str, prefix: str) -> urls.URL:
    if prefix:
        prefix = prefix.rstrip("/")

    return urls.URL(f"s3://{bucket}/{prefix}")



"""Postgres query client.

Adds the convinience methods for loading csv data using copy_expert,
which is more performant than using sql alchemy functions.
"""

import io
from typing import Sequence

import pandas as pd

from tentaclio import protocols

from . import decorators, sqla_client


__all__ = ["PostgresClient"]


class PostgresClient(sqla_client.SQLAlchemyClient):
    """Postgres client, backed by a SQLAlchemy connection."""

    allowed_schemes = ["postgresql"]

    # Postgres Copy Expert methods:
    @decorators.check_conn
    def get_df_unsafe(self, sql_query: str, params: dict = None, **kwargs) -> pd.DataFrame:
        """Run a raw SQL query and return a data framem using COPY.

        Params:
            sql_query: query to execute
            params: not supported; must be None
            **kwargs: additional kwargs to pass to `pandas.read_csv`
        """
        if params:
            raise NotImplementedError(
                "Support for `params` is not implemented; please provide a pre-formatted query."
            )
        copy_sql = f"COPY ({sql_query.rstrip(';')}) TO STDOUT WITH CSV HEADER"
        buffer = io.StringIO()
        raw_conn = self._get_raw_conn()
        raw_conn.cursor().copy_expert(copy_sql, buffer)
        buffer.seek(0)
        df = pd.read_csv(buffer, **kwargs)
        return df

    @decorators.check_conn
    def dump_df(self, df: pd.DataFrame, dest_table: str) -> None:
        """Dump a data frame into an existing Postgres table."""
        buff = io.StringIO()
        df.to_csv(buff, index=False)
        buff.seek(0)
        self._copy_expert_csv(buff, df.columns, dest_table)

    @decorators.check_conn
    def dump_csv(
        self, csv_reader: protocols.Reader, columns: Sequence[str], dest_table: str
    ) -> None:
        """Dump a csv reader into the database."""
        self._copy_expert_csv(csv_reader, columns, dest_table)

    def _copy_expert_csv(
        self, csv_reader: protocols.Reader, columns: Sequence[str], dest_table: str
    ) -> None:
        """Dump a csv reader into the given table."""
        sql_columns = ",".join(columns)
        sql_query = f"""COPY {dest_table} ({sql_columns}) FROM STDIN
                        WITH CSV HEADER DELIMITER AS ','
                        NULL AS 'NULL';"""
        raw_conn = self._get_raw_conn()
        try:
            raw_conn.cursor().copy_expert(sql_query, csv_reader)
        except Exception:
            raw_conn.rollback()
            raise
        else:
            raw_conn.commit()
